﻿Public Class MazeForm

    Dim Maze As Maze
    Dim Drawer As Painter
    Dim Buffer As BufferedGraphics
    Dim Origin As New Point(16, 16)
    Dim Solution As New List(Of Point)
    Dim SizeBound As Boolean
    Dim LastFile As String

    ' Positioning of maze
    Public MazeBoundary As New Rectangle

    Dim StartNode As New Point(0, 0)
    Dim EndNode As New Point(0, 0)

    Private Sub NewMazeToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles NewMazeToolStripMenuItem.Click

        If NewMazeDialog.ShowDialog = Windows.Forms.DialogResult.OK Then
            LastFile = ""
            SolveToolStripMenuItem.Checked = False
            NewMaze()
        End If

    End Sub

    Private Sub SolveToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles SolveToolStripMenuItem.Click

        Solution.Clear()

        ' See if we need to solve the maze or not
        If MazeLoaded And Not SolveToolStripMenuItem.Checked Then
            If Maze.FindPath(StartNode, EndNode, Solution) Then
                SolveToolStripMenuItem.Checked = True
                Drawer.DrawPath(Buffer.Graphics, MazeBoundary.Location, Solution)
                Me.Refresh()
            Else
                MessageBox.Show("Couldn't find exit node!", "Error")
            End If
        Else
            UpdateBuffer()
            Me.Refresh()
            SolveToolStripMenuItem.Checked = False
        End If

    End Sub

    Private Sub UpdateBuffer()

        ' Calculates the neccessary size of the buffer
        Dim newSize = Drawer.ScreenSize(Maze)

        ' Allocate a new buffer if we must
        If Buffer Is Nothing OrElse MazeBoundary.Size <> newSize Then
            CleanUp() : MazeBoundary.Size = newSize
            Buffer = BufferedGraphicsManager.Current.Allocate(Me.CreateGraphics,
                                                              New Rectangle(0, 0, MazeBoundary.Right, MazeBoundary.Bottom))
        End If

        ' Then, draw the maze onto the buffer and update the form
        DrawMaze(Buffer.Graphics, Drawer, Maze, Me.BackColor, MazeBoundary.Location)

    End Sub

    Private Sub DrawMaze(ByVal Surface As Graphics, ByVal Converter As Painter, ByVal Item As Maze,
                         ByVal Background As Color, ByVal Offset As Point)

        ' Draw unto any surface
        Surface.Clear(Background)
        Converter.DrawMaze(Surface, Item, Offset)

    End Sub

    Private Sub NewMaze()

        ' Firstly, initialize graphics object
        Drawer = New Painter With {.CellSize = New Size(NewMazeDialog.CellWidthTextBox.Text,
                                                        NewMazeDialog.CellHeightTextBox.Text)}
        SizeBound = NewMazeDialog.UsePrintSizeCheck.Checked

        ' Fix pen
        If Drawer.CellSize.Width < 4 OrElse Drawer.CellSize.Height < 4 Then
            Drawer.Solution = Pens.Red
        End If

        ' Then, allocate maze
        If SizeBound Then
            Maze = FitMazeToPrinter()
        Else
            Maze = New Maze(NewMazeDialog.ColumnsTextBox.Text, NewMazeDialog.RowsTextBox.Text)
        End If

        EndNode = New Point(Maze.Columns - 1, Maze.Rows - 1)
        Maze.Generate(New Point(0, 0))
        Maze(StartNode).Corner(Sides.Top) = False
        Maze(EndNode).Corner(Sides.Right) = False
        UpdateBuffer()
        Me.Refresh()

    End Sub

    Private Sub CleanUp()

        ' Clears the buffer if it's set
        If Buffer IsNot Nothing Then
            Buffer.Dispose() : Buffer = Nothing
        End If

    End Sub

    Private Sub SaveMaze(ByVal Path As String, ByVal Converter As Painter)

        Dim newSize = Converter.ScreenSize(Maze)

        Using Temporary As New Bitmap(newSize.Width, newSize.Height)

            ' Store the geography of the map
            Using Surface = Graphics.FromImage(Temporary)
                DrawMaze(Surface, Converter, Maze, Color.White, Point.Empty)
            End Using

            Temporary.Save(Path, Imaging.ImageFormat.Bmp)
        End Using

    End Sub

    Private Function FitMazeToPrinter() As Maze

        ' Create appropriate maze
        With PageSetup.PageSettings
            Dim Size As New Size((.PrintableArea.Width - .Margins.Left - .Margins.Right) \ Drawer.CellSize.Width,
                                 (.PrintableArea.Height - .Margins.Top - .Margins.Bottom) \ Drawer.CellSize.Height)
            Return If(.Landscape, New Maze(Size.Height, Size.Width), New Maze(Size.Width, Size.Height))
        End With

    End Function

    Private Sub PrintMaze_PrintPage(ByVal sender As System.Object, ByVal e As System.Drawing.Printing.PrintPageEventArgs) Handles PrintMaze.PrintPage

        ' ToDo: Handle large mazes
        Drawer.DrawMaze(e.Graphics, Maze, Point.Empty)
        If SolveToolStripMenuItem.Checked Then Drawer.DrawPath(e.Graphics, Point.Empty, Solution)

    End Sub

    Private Sub PageSetupToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles PageSetupToolStripMenuItem.Click
        PageSetup.Document = PrintMaze
        If PageSetup.ShowDialog() AndAlso SizeBound Then
            If FitMazeToPrinter().Dimensions <> Maze.Dimensions Then
                If MessageBox.Show("Document size has changed. Do you want to change the size of the maze as well?", "Change",
                                   MessageBoxButtons.YesNo, MessageBoxIcon.Question) = Windows.Forms.DialogResult.Yes Then
                    SolveToolStripMenuItem.Checked = False
                    NewMaze()
                End If
            End If
        End If
    End Sub

    Private Sub SaveToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles SaveToolStripMenuItem.Click
        If LastFile.Length > 0 Then
            SaveMaze(LastFile, Drawer)
        Else
            SaveAsToolStripMenuItem_Click(sender, e)
        End If
    End Sub

    Private Sub SaveAsToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles SaveAsToolStripMenuItem.Click
        If SaveDialog.ShowDialog = Windows.Forms.DialogResult.OK Then
            SaveMaze(SaveDialog.FileName, Drawer)
        End If
    End Sub

#Region "Trivial"
    Public ReadOnly Property MazeLoaded() As Boolean
        Get
            Return Maze IsNot Nothing
        End Get
    End Property

    Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
        If Buffer IsNot Nothing Then Buffer.Render(e.Graphics)
    End Sub

    Private Sub MazeForm_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        MazeBoundary = New Rectangle(Origin.X, Origin.Y + MainMenu.Bottom, 0, 0)
        PrintMaze.DefaultPageSettings.Landscape = True
    End Sub

    Private Sub ExitToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ExitToolStripMenuItem.Click
        Me.Close()
    End Sub

    Private Sub PrintToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles PrintToolStripMenuItem.Click
        If PrintDialog.ShowDialog() = Windows.Forms.DialogResult.OK Then
            PrintMaze.Print()
        End If
    End Sub

    Private Sub FileToolStripMenuItem_DropDownOpening(ByVal sender As Object, ByVal e As System.EventArgs) Handles FileToolStripMenuItem.DropDownOpening
        SaveToolStripMenuItem.Enabled = Maze IsNot Nothing
        SaveAsToolStripMenuItem.Enabled = SaveToolStripMenuItem.Enabled
        PrintToolStripMenuItem.Enabled = SaveToolStripMenuItem.Enabled
    End Sub

    Private Sub MazeToolStripMenuItem_DropDownOpened(ByVal sender As Object, ByVal e As System.EventArgs) Handles MazeToolStripMenuItem.DropDownOpened
        SolveToolStripMenuItem.Enabled = Maze IsNot Nothing
    End Sub

    Protected Overrides Sub Finalize()
        CleanUp()
        MyBase.Finalize()
    End Sub
#End Region
End Class
